#include "config.h"
#include <errno.h>
#include <time.h>
#include <stdlib.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <oniguruma.h>
#include <sys/stat.h>
#include <signal.h>
#include <limits.h>
#include <dirent.h>

#if defined(HAVE_CURSES_H)
#include <curses.h>
#elif defined(HAVE_NCURSES_H)
#include <ncurses.h>
#elif defined(HAVE_NCURSES_NCURSES_H)
#include <ncurses/ncurses.h>
#endif

#include "pipes/pipes.h"

BOOL cmd_jobs(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    char buf[BUFSIZ];
    int i;
    for(i=0; i<vector_size(gJobs); i++) {
        sObject* job = vector_item(gJobs, i);
        snprintf(buf, BUFSIZ, "[%d] name: %s pid&pgrp: %d", i+1, SJOB(job).mName, SJOB(job).mPGroup);
        if(!fd_write(nextout, buf)) {
            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
            return FALSE;
        }

        if(!fd_write(nextout, "\n")) {
            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
            return FALSE;
        }
    }
    runinfo->mRCode = 0;
    return TRUE;
}

BOOL cmd_fg(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    if(gAppType != kATOptC) {
        if(runinfo->mCommand->mArgsNumRuntime >= 2) {
            int num = atoi(runinfo->mCommand->mArgsRuntime[1]);
            if(!forground_job(num-1)) {
                return FALSE;
            }
        }
        else {
            if(!forground_job(0)) {
                return FALSE;
            }
        }

        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_exit(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;
    
    if(sCommand_option_item(command, "-force")) {
        pipes_kill_all_jobs();

        runinfo->mRCode = RCODE_EXIT;
    }
    else {
        if(vector_size(gJobs) > 0) {
            err_msg("jobs exist", runinfo->mSName, runinfo->mSLine);
            return FALSE;
        }

        runinfo->mRCode = RCODE_EXIT;
    }

    return FALSE;
}

BOOL cmd_while(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;
    if(command->mBlocksNum == 2) {
        while(1) {
            /// condition ///
            int rcode;
            if(!run(command->mBlocks[0], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                runinfo->mRCode = rcode;
                return FALSE;
            }

            if(rcode != 0) {
                break;
            }

            /// inner loop ///
            if(!run(command->mBlocks[1], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                if(rcode == RCODE_BREAK) {
                    break;
                }
                else {
                    runinfo->mRCode = rcode;
                    return FALSE;
                }
            }
        }

        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_break(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    runinfo->mRCode = RCODE_BREAK;
    return FALSE;
}

BOOL cmd_true(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    runinfo->mRCode = 0;
    return TRUE;
}

BOOL cmd_false(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    runinfo->mRCode = RCODE_NFUN_FALSE;
    return TRUE;
}

BOOL cmd_if(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mBlocksNum > 1) {
        if(command->mBlocksNum % 2 == 0) {
            int i;
            for(i=0; i<command->mBlocksNum; i+=2) {
                /// a condition ///
                int rcode = 0;
                if(!run(command->mBlocks[i], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                    runinfo->mRCode = rcode;
                    return FALSE;
                }

                /// a statment ///
                if(rcode == 0) {
                    int rcode = 0;
                    if(!run(command->mBlocks[i+1], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                        runinfo->mRCode = rcode;
                        return FALSE;
                    }
                    break;
                }
            }
        }
        else {
            BOOL run_else = TRUE;

            int i;
            for(i=0; i<command->mBlocksNum-1; i+=2) {
                /// a condition ///
                int rcode = 0;
                if(!run(command->mBlocks[i], nextin, nextout,  &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                    runinfo->mRCode = rcode;
                    return FALSE;
                }

                /// a statment ///
                if(rcode == 0) {
                    run_else = FALSE;

                    int rcode = 0;
                    if(!run(command->mBlocks[i+1], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                        runinfo->mRCode = rcode;
                        return FALSE;
                    }
                    break;
                }
            }

            /// else statment ///
            if(run_else) {
                int rcode = 0;
                if(!run(command->mBlocks[command->mBlocksNum-1], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                    runinfo->mRCode = rcode;
                    return FALSE;
                }
            }
        }

        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_return(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    runinfo->mRCode = RCODE_RETURN;
    return FALSE;
}

BOOL cmd_stackinfo(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    char buf[BUFSIZ];

    snprintf(buf, BUFSIZ, "slot size %d\npool size %d\n", stack_slot_size(), stack_page_size());

    if(!fd_write(nextout, buf)) {
        err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
        runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
        return FALSE;
    }

    runinfo->mRCode = 0;
    return TRUE;
}

BOOL cmd_stackframe(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    char buf[BUFSIZ];

    hash_it* it = hash_loop_begin(SFUN(runinfo->mRunningObject).mLocalObjects);
    while(it) {
        sObject* object2 = hash_loop_item(it);
        char* key = hash_loop_key(it);

        char*obj_kind[16] = {
            NULL, "var", "array", "hash", "list", "native function", "block", "file dicriptor", "job", "object", "memchecker", "function", "class", NULL, NULL, NULL
        };
        snprintf(buf, BUFSIZ, "%s: %s\n", key, obj_kind[TYPE(object2)]);

        if(!fd_write(nextout, buf)) {
            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
            return FALSE;
        }

        it = hash_loop_next(it);
    }

    runinfo->mRCode = 0;
    return TRUE;
}

BOOL cmd_gcinfo(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    char buf[BUFSIZ];

    snprintf(buf, BUFSIZ, "free objects %d\nused objects %d\nall object slot %d\nslot size %d\npool size %d\n", gc_free_objects_num(), gc_used_objects_num(), gc_slot_size()*gc_pool_size(), gc_slot_size(), gc_pool_size());

    if(!fd_write(nextout, buf)) {
        err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
        runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
        return FALSE;
    }

    runinfo->mRCode = 0;
    return TRUE;
}

BOOL cmd_subshell(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mBlocksNum == 1) {
        int rcode = 0;
        if(!run(command->mBlocks[0], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
            runinfo->mRCode = rcode;
            return FALSE;
        }
        runinfo->mRCode = rcode;
    }
    
    return TRUE;
}

BOOL cmd_sweep(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mArgsNumRuntime == 1) {
        int num = gc_sweep();

        char buf[BUFSIZ];
        snprintf(buf, BUFSIZ, "%d objects deleted\n", num);

        if(!fd_write(nextout, buf)) {
            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
            return FALSE;
        }
        runinfo->mRCode = 0;
    }
    else {
        int i;
        for(i=1; i<command->mArgsNumRuntime; i++) {
            char* var_name = command->mArgsRuntime[i];
            sObject* current_object = runinfo->mCurrentObject;

            sObject* item = hash_item(current_object, var_name);

            if(item->mUserObject) {
                if(hash_erase(current_object, var_name)) {
                    runinfo->mRCode = 0;
                }
            }
        }
    }

    return TRUE;
}

BOOL cmd_print(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    sObject* out;
    if(sCommand_option_item(command, "-error")) {
        out = gStderr;
    }
    else {
        out = nextout;
    }

    if(command->mArgsNumRuntime > 1) {
        int i;
        for(i=1; i<command->mArgsNumRuntime-1; i++) {
            if(!fd_write(out, command->mArgsRuntime[i])) {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }
            if(!fd_write(out, " ")) {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }
        }
        if(!fd_write(out, command->mArgsRuntime[i])) {
            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
            return FALSE;
        }
        runinfo->mRCode = 0;
    }
    else if(runinfo->mFilter) {
        if(sCommand_option_item(command, "-read-from-error")) {
            if(!fd_write(out, SFD(gStderr).mBuf)) {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }
            fd_clear(gStderr);
        }
        else {
            if(!fd_write(out, SFD(nextin).mBuf)) {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }
        }
        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_load(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;
    if(command->mArgsNumRuntime >= 2) {
        runinfo->mRCode = 0;
        if(!load_file(command->mArgsRuntime[1], nextin, nextout, runinfo, command->mArgsRuntime + 2, command->mArgsNumRuntime -2))
        {
            return FALSE;
        }
    }

    return TRUE;
}

BOOL cmd_eval(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;
    if(command->mBlocksNum > 0) {
        int rcode;
        if(!run(command->mBlocks[0], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
            runinfo->mRCode = rcode;
            return FALSE;
        }
        runinfo->mRCode = 0;
    }
    else if(command->mArgsNumRuntime >= 2) {
        char* buf = command->mArgsRuntime[1];

        stack_start_stack();
        int rcode = 0;
        sObject* block = BLOCK_NEW_STACK();
        int sline = 1;
        if(parse(buf, "eval", &sline, block, NULL)) {
            if(!run(block, nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                runinfo->mRCode = rcode;
                stack_end_stack();
                return FALSE;
            }
        }
        else {
            err_msg("parser error\n", runinfo->mSName, runinfo->mSLine);
            stack_end_stack();
            return FALSE;
        }

        stack_end_stack();
        runinfo->mRCode = 0;
    }
    else if(runinfo->mFilter) {
        char* buf = SFD(nextin).mBuf;

        stack_start_stack();
        int rcode = 0;
        sObject* block = BLOCK_NEW_STACK();
        int sline = 1;
        if(parse(buf, "eval", &sline, block, NULL)) {
            if(!run(block, nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                runinfo->mRCode = rcode;
                stack_end_stack();
                return FALSE;
            }
        }
        else {
            err_msg("parser error\n", runinfo->mSName, runinfo->mSLine);
            stack_end_stack();
            return FALSE;
        }

        stack_end_stack();
        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_msleep(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mArgsNumRuntime == 2) {
        ///  ///
        char c [] = {'.', 'o', 'O'};
        int n = atoi(command->mArgsRuntime[1]);
        int i;
        for(i=0; i < n*5; i++) {
            char buf[32];
            snprintf(buf, 32, "%c\033[D", c[i % 3]);

            if(!fd_write(nextout, buf)) {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }
            if(nextout == gStdout) {
                if(!fd_flash(nextout, 1)) {
                    err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                    runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                    return FALSE;
                }
            }

            usleep(200000);
        }

        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_raise(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mArgsNumRuntime == 2) {
        err_msg(command->mArgsRuntime[1], runinfo->mSName, runinfo->mSLine);
        return FALSE;
    }

    return TRUE;
}


BOOL cmd_time(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mBlocksNum > 0) {
        time_t timer;
        time(&timer);

        int rcode;
        if(!run(command->mBlocks[0], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
            runinfo->mRCode = rcode;
            return FALSE;
        }
        runinfo->mRCode = 0;

        time_t timer2;
        time(&timer2);
       
        int sec = timer2 - timer;
        int minuts = sec / 60;
        int sec2 = sec - minuts * 60;

        char buf[1024];
        snprintf(buf, 1024, "%d sec(%d minuts %d sec)\n", sec, minuts, sec2);

        if(!fd_write(nextout, buf) ) {
            err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
            runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
            return FALSE;
        }
    }

    return TRUE;
}

BOOL cmd_umask(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(!runinfo->mFilter && command->mArgsNumRuntime == 2) {
        int number = strtol(command->mArgsRuntime[1], NULL, 8);
        umask(number);
        runinfo->mRCode = 0;
    }

    return TRUE;
}

static BOOL pipes_seek_external_programs(char* path, sObject* sys, char* sname, int sline)
{
    DIR* dir = opendir(path);
    if(dir) {
        struct dirent* entry;
        while(entry = readdir(dir)) {
            char path2[PATH_MAX];
            snprintf(path2, PATH_MAX, "%s/%s", path, entry->d_name);
            
            struct stat stat_;
            memset(&stat_, 0, sizeof(struct stat));
            stat(path2, &stat_);

            if(strcmp(entry->d_name, ".") != 0
                && strcmp(entry->d_name, "..") != 0
                &&
                (stat_.st_mode & S_IXUSR ||stat_.st_mode & S_IXGRP||stat_.st_mode & S_IXOTH)
                && hash_item(sys, entry->d_name) == NULL)
            {
                hash_put(sys, entry->d_name, EXTERNAL_NEW_GC(path2, TRUE));
            }
        }

        closedir(dir);
    }
    else {
        char buf[BUFSIZ];
        snprintf(buf, BUFSIZ, "%s is invalid path", path);
        err_msg(buf, sname, sline);
        return FALSE;
    }

    return TRUE;
}

BOOL pipes_rehash(char* sname, int sline)
{
    char* path = getenv("PATH");
    if(path == NULL) {
        err_msg("PATH environment variables is NULL", sname, sline);
        return FALSE;
    }

    sObject* sys = UOBJECT_NEW_GC(8, gRootObject, "sys", FALSE);
    hash_put(gRootObject, "sys", sys);
    uobject_init(sys);

    char buf[PATH_MAX+1];

    char* p = path;
    while(1) {
        char* p2;
        if(p2 = strstr(p, ":")) {
            int size = p2 - p;
            if(size > PATH_MAX) {
                size = PATH_MAX;
            }
            memcpy(buf, p, size);
            buf[size] = 0;

            p = p2 + 1;

            if(!pipes_seek_external_programs(buf, sys, sname, sline)) {
                return FALSE;
            }
        }
        else {
            strncpy(buf, p, PATH_MAX);

            if(!pipes_seek_external_programs(buf, sys, sname, sline)) {
                return FALSE;
            }
            break;
        }
    }

    return TRUE;
}

BOOL cmd_rehash(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    if(!runinfo->mFilter) {
        if(!pipes_rehash(runinfo->mSName, runinfo->mSLine)) {
            return FALSE;
        }
        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_try(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;

    if(command->mBlocksNum == 2) {
        int rcode = 0;
        if(!run(command->mBlocks[0], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
            if(rcode == RCODE_NFUN_FALSE || rcode == RCODE_BREAK || rcode == RCODE_RETURN || rcode == RCODE_EXIT) {
                runinfo->mRCode = rcode;
                return FALSE;
            }
            else {
                if(!run(command->mBlocks[1], nextin, nextout, &rcode, runinfo->mCurrentObject, runinfo->mRunningObject)) {
                    runinfo->mRCode = rcode;
                    return FALSE;
                }
                else {
                    runinfo->mRCode = 0;
                    return TRUE;
                }
            }
        }
        else {
            runinfo->mRCode = 0;
            return TRUE;
        }
    }

    return TRUE;
}

BOOL cmd_errmsg(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    if(!fd_write(nextout, string_c_str(gErrMsg))) {
        err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
        runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
        return FALSE;
    }
    runinfo->mRCode = 0;

    return TRUE;
}

sObject* gPrompt;

BOOL cmd_prompt(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;
    if(command->mBlocksNum == 1) {
        gPrompt = block_clone_gc(command->mBlocks[0], T_BLOCK, FALSE);
        hash_put(gPipesObject, "_prompt", gPrompt);
        runinfo->mRCode = 0;
    }

    return TRUE;
}

BOOL cmd_block(sObject* nextin, sObject* nextout, sRunInfo* runinfo)
{
    sCommand* command = runinfo->mCommand;
    sObject* fun = runinfo->mRunningObject;

    if(sCommand_option_item(command, "-number")) {
        if(SFUN(fun).mArgBlocks) {
            char buf[32];
            snprintf(buf, 32, "%d\n", vector_size(SFUN(fun).mArgBlocks));

            if(!fd_write(nextout, buf))
            {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }

            runinfo->mRCode = 0;
        }
    }
    else if(sCommand_option_item(command, "-run")) {
        int block_num;
        if(command->mArgsNumRuntime > 1) {
            block_num = atoi(command->mArgsRuntime[1]);
        }
        else {
            block_num = 0;
        }

        if(SFUN(fun).mArgBlocks && block_num < vector_size(SFUN(fun).mArgBlocks)) {
            sObject* block = vector_item(SFUN(fun).mArgBlocks, block_num);
            sObject* fun = fun_clone_from_stack_block_stack(block, NULL);

            if(!run_function(fun, runinfo->mCurrentObject, nextin, nextout, runinfo, command->mArgsRuntime + 2, command->mArgsNumRuntime -2, command->mBlocks, command->mBlocksNum))
            {
                return FALSE;
            }

            runinfo->mRCode = 0;
        }
    }
    else {
        int block_num;
        if(command->mArgsNumRuntime > 1) {
            block_num = atoi(command->mArgsRuntime[1]);
        }
        else {
            block_num = 0;
        }

        if(SFUN(fun).mArgBlocks && block_num < vector_size(SFUN(fun).mArgBlocks)) {
            sObject* block = vector_item(SFUN(fun).mArgBlocks, block_num);

            if(!fd_write(nextout, SBLOCK(block).mSource))
            {
                err_msg("signal interrupt", runinfo->mSName, runinfo->mSLine);
                runinfo->mRCode = RCODE_SIGNAL_INTERRUPT;
                return FALSE;
            }

            runinfo->mRCode = 0;
        }
    }

    return TRUE;
}
